import os
from urllib.request import urlopen

import numpy as np

import astropy.units as u
import astropy.table as atable
from astropy.time import Time
from astropy.coordinates import SkyCoord, Angle, AltAz, EarthLocation

from casacore import tables

import lsmtool
from lsmtool import tableio
lsmtool.logger.setLevel('warning')

from .settings import TEMPLATE_DIR


# Lazy loaded:
# - astroquery.vizier in build_sky_model_from_specfind()
# - nenupy.beam in compute_beam()


all_mas = [0, 1, 3, 7, 11, 13]

nenufar_location = EarthLocation(lat=47.376511 * u.deg, lon=2.1924002 * u.deg)

sky_model_catalogs = ['specfind', 'lcs165']


def get_optfreq(fmhz):
    if fmhz <= 50:
        return 40
    if fmhz <= 68:
        return 60
    return 80

def get_lowres_ateam_skymodel_path():
    return os.path.join(TEMPLATE_DIR, 'Ateam_lowres.skymodel')


def compute_beam(altaz, altaz_phase_center, fmhz=70, ma=0, pol='NW'):
    from nenupy.beam import ABeam

    anabeam = ABeam(fmhz * u.MHz, pol, altaz_phase_center.az.deg, altaz_phase_center.alt.deg, ma=ma)
    anabeam.squint_freq = get_optfreq(fmhz)

    return anabeam.beam_values(altaz.transform_to('icrs'), Time(altaz.obstime.datetime))


def compute_beam_freqs(altaz, altaz_phase_center, fmhz, ma=0, pol='NW'):
    return np.array([compute_beam(altaz, altaz_phase_center, ma=ma, fmhz=k, pol='NW') for k in fmhz])


def compute_beam_freqs_mas(altaz, altaz_phase_center, fmhz, mas, pol='NW'):
    return np.array([compute_beam_freqs(altaz, altaz_phase_center, fmhz, ma=k, pol='NW') for k in mas])


def build_sky_model_from_specfind(coord, min_flux, radius, out_file, waste_spidx=-0.8):
    from astroquery.vizier import Vizier
    Vizier.ROW_LIMIT = 100000

    print(f'Querying SPECFIND v2 spec catalog for phase center {coord}...')

    catalog = Vizier(catalog='VIII/85A/spectra', row_limit=100000,
                     column_filters={'nu': '30..200', 'S(nu)': f'> {0.2 * 0.5 * 1e3}',
                                     'Name': '~|4C*|6C*|7C*|8C*|VLSS*'})

    table = catalog.query_region(coord, radius=Angle(radius, "deg"))[0]
    # Get unique physical source
    tg = table.group_by('Seq')
    table_uniq = tg[tg.groups.indices[:-1]]

    # Compute flux at 74 Mhz from the spectra index
    table_uniq['S_nu_74'] = (10 ** (table_uniq['a'] * np.log10(74) + table_uniq['b']))
    table_uniq = table_uniq[table_uniq['S_nu_74'] > min_flux * 1e3]

    coords = SkyCoord(ra=table_uniq['RAJ2000'], dec=table_uniq['DEJ2000'])

    print(f'-> Found {len(table_uniq)} sources with flux > {min_flux} Jy with radius > {radius} deg')

    print('Querying SPECFIND v2 waste catalog ...')
    catalog = Vizier(catalog='VIII/85A/waste', row_limit=100000,
                     column_filters={'nu': '30..200', 'S(nu)': f'> {0.2 * 0.5 * 1e3}',
                                     'Name': '~|4C*|6C*|7C*|8C*|VLSS*'})

    table_waste = catalog.query_region(coord, radius=Angle(radius, "deg"))[0]
    table_waste = table_waste[table_waste['Seq'] < 0]
    # Get unique physical source
    tg = table_waste.group_by('Seq')
    table_waste = tg[tg.groups.indices[:-1]]
    # Look for duplicate name entries
    tg = table_waste.group_by('Name')
    table_waste = tg[tg.groups.indices[:-1]]
    table_waste['S_nu_74'] = table_waste['S_nu_']

    coords_waste = SkyCoord(ra=table_waste['RAJ2000'], dec=table_waste['DEJ2000'])

    # Extract the extra sources which are not in the main table
    _, idx2, _, _ = coords_waste.search_around_sky(coords, 0.2 * u.deg)
    table_extra = table_waste[np.setdiff1d(np.arange(len(coords_waste)), idx2)]

    print(f'-> Found {len(table_extra)} extra sources in waste table (default spectra index: {waste_spidx})')

    table_merge = atable.vstack([table_uniq, table_extra], metadata_conflicts='silent')
    table_merge['a'].fill_value = waste_spidx
    table_merge.sort('S_nu_74', reverse=True)

    coords_merge = SkyCoord(ra=table_merge['RAJ2000'], dec=table_merge['DEJ2000'])

    print(f'Building sky model with {len(table_merge)} sources ...')

    sky_model = atable.Table(names=['Name', 'Type', 'Patch', 'Ra', 'Dec', 'I', 'ReferenceFrequency', 'SpectralIndex'],
                             dtype=[str] * 8, data=np.zeros((8, len(table_merge))).T)
    sky_model['Ra'] = coords_merge.ra.deg
    sky_model['Ra'].unit = u.deg

    sky_model['Dec'] = coords_merge.dec.deg
    sky_model['Dec'].unit = u.deg

    sky_model['Name'] = table_merge['Name']
    sky_model['Patch'] = 'Main'
    sky_model['Type'] = 'Point'
    sky_model['ReferenceFrequency'] = [74e6] * len(table_merge)
    sky_model['ReferenceFrequency'].unit = u.Hz

    sky_model['I'] = table_merge['S_nu_74'] * 1e-3
    sky_model['I'].unit = u.Jansky

    sky_model['SpectralIndex'] = [np.array([a], dtype=object) for a in table_merge['a'].filled()]

    print(f'Saving Intrinsic sky model to {out_file} ...')

    tableio.skyModelWriter(sky_model, out_file)

    return lsmtool.load(out_file)


def build_sky_model_from_lcs165(coord, min_flux, radius, out_file):
    assert radius <= 20, 'Radius is limited to 20 degrees for lcs165 sky models'
    url = 'https://lcs165.lofar.eu/cgi-bin/gsmv1.cgi'
    cmd = f'{url}?coord={coord.ra.deg},{coord.dec.deg}&radius={radius}&cutoff={min_flux}&unit=deg&deconv=y'
    print('Sending request:', cmd)
    with urlopen(cmd) as response:
        response = response.read()

    print(f'Saving Intrinsic sky model to {out_file} ...')

    with open(out_file, mode='w') as fp:
        fp.write(response.decode('utf-8'))

    sky_model = lsmtool.load(out_file)
    set_patch_name(sky_model, 'Main')
    sky_model.write(out_file, clobber=True)

    print(f'Sky model saved with {len(sky_model)} components')

    return sky_model


def build_sky_model(coord, min_flux, radius, out_file, catalog='specfind'):
    if catalog == 'specfind':
        sky_model = build_sky_model_from_specfind(coord, min_flux, radius, out_file)
    else:
        sky_model = build_sky_model_from_lcs165(coord, min_flux, radius, out_file)

    return sky_model


def build_sky_model_ms(ms_file, min_flux, radius, out_file, catalog='specfind'):
    phase_dir = tables.table(ms_file + '/FIELD').getcol('PHASE_DIR').squeeze()
    coord_phase_dir = SkyCoord(ra=phase_dir[0] * u.rad, dec=phase_dir[1] * u.rad)

    return build_sky_model(coord_phase_dir, min_flux, radius, out_file, catalog=catalog)


def print_info(model):
    patch_count = dict(zip(*np.unique(model.getColValues('Patch'), return_counts=True)))
    if model.getPatchNames() is not None:
        for patch, tot_flux in zip(model.getPatchNames(), model.getColValues('I', aggregate='sum')):
            print(f'- {patch}: {patch_count[patch]} cmpts totaling {tot_flux:.1f} Jy')


def apply_nenufar_beam(sky_model, observing_time, coord, mfmhz):
    print('Sky model info:')
    print_info(sky_model)

    print(f'Computing beam:')
    if isinstance(observing_time.iso, np.ndarray):
        print(f'--Dates: {" ".join(observing_time.iso)}')
    else:
        print(f'--Date: {observing_time.iso}')
    print(f'--Pointing: {coord.to_string("hmsdms")}')
    print(f'--Frequency: {mfmhz} MHz')

    beams = []

    for obs_time in observing_time:
        altaz = AltAz(location=nenufar_location, obstime=obs_time)

        altaz_phase_dir = coord.transform_to(altaz)

        coord_skymodel = SkyCoord(ra=sky_model.getColValues('Ra') * u.deg, dec=sky_model.getColValues('Dec') * u.deg)
        altaz_skymodel = coord_skymodel.transform_to(altaz)

        beam_nw = compute_beam_freqs_mas(altaz_skymodel, altaz_phase_dir, [mfmhz], all_mas, pol='NW')
        beam_ne = compute_beam_freqs_mas(altaz_skymodel, altaz_phase_dir, [mfmhz], all_mas, pol='NE')
        beam = 0.5 * (beam_nw + beam_ne)
        beam_pointing = compute_beam(SkyCoord([altaz_phase_dir]), altaz_phase_dir, fmhz=mfmhz)
        beam = beam.squeeze()
        # print(beam.max(axis=1), beam_pointing.max(axis=1))
        beam = beam / beam_pointing.squeeze()
        beam[np.isnan(beam)] = 0
        beams.append(beam)

    beams = np.array(beams)

    # Take the mean of the beam of all MA rotations
    app_sky_model = sky_model.copy()
    flux = app_sky_model.getColValues('I') * beams.mean(axis=1).max(axis=0).squeeze()
    app_sky_model.setColValues('I', flux)

    print('Sky model info after applying beam:')
    print_info(app_sky_model)

    return app_sky_model


def edit_model(sky_model, observing_time, min_flux=0.5, min_flux_patch=20, min_elevation_patch=10, 
               always_keep=['Main'], always_remove=[]):
    print(f'Removing components with flux < {min_flux} Jy ...')
    sky_model.remove(f'I < {min_flux}')

    print(f'Removing patch with total flux < {min_flux_patch} Jy ...')

    for patch, sum_I in zip(sky_model.getPatchNames(), sky_model.getColValues('I', aggregate='sum')):
        if sum_I < min_flux_patch and ((patch not in always_keep) or sum_I < 1):
            print(f'-- Removing patch {patch}')
            sky_model.remove(f'Patch == {patch}')

    if min_elevation_patch > 0:
        print(f'Removing patch with elevation < {min_elevation_patch} deg ...')
        altaz = AltAz(location=nenufar_location, obstime=observing_time)

        for patch, (ra, dec) in sky_model.getPatchPositions(method='wmean').items():
            elevation = SkyCoord(ra=ra.deg * u.deg, dec=dec.deg * u.deg).transform_to(altaz).alt.max()
            print(f'-- Max elevation of {patch}: {elevation:.2f}')
            if (np.isnan(elevation.deg) or elevation.deg < min_elevation_patch) \
                    and ((patch not in always_keep) or elevation < 0):
                print(f'-- Removing Patch {patch} (elevation {elevation.deg} deg)')
                sky_model.remove(f'Patch == {patch}')

    if always_remove is not None:
        print(f'Removing user specified patch ...')
        for patch in sky_model.getPatchNames():
            if patch in always_remove:
                print(f'-- Removing Patch {patch}')
                sky_model.remove(f'Patch == {patch}')

    print('Sky model info after model editing:')
    print_info(sky_model)

    return sky_model


def add_mssing_logarithmic_si_col(skymodel):
    if 'LogarithmicSI' not in skymodel.getColNames():
        skymodel.setColValues('LogarithmicSI', ['true'] * len(skymodel))


def concatenate(sky_model_filenames):
    sky_model = lsmtool.load(sky_model_filenames[0])
    add_mssing_logarithmic_si_col(sky_model)
    for f in sky_model_filenames[1:]:
        other_sky_model = lsmtool.load(f)
        add_mssing_logarithmic_si_col(other_sky_model)
        lsmtool.operations.concatenate.concatenate(sky_model, other_sky_model)

    return sky_model


def set_patch_name(sky_model, patch_name):
    sky_model.setColValues('PATCH', [patch_name] * len(sky_model), index=2)
    sky_model.setPatchPositions(method='wmean')



from unittest import mock

import numpy as np

from nenucal import msutils


class MockTable(object):

    def __init__(self, data, keywords):
        self.data = data
        self.keywords = keywords

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def getcol(self, col):
        return self.data[col]

    def getcell(self, col, i):
        return self.data[col][i]

    def nrows(self):
        return self.data[list(self.data.keys())[0]].shape[0]

    def putcol(self, col, newdata):
        self.data[col] = newdata

    def getkeyword(self, col):
        return self.keywords[col]

    def close(self):
        pass


class MockTables(object):

    def __init__(self):
        self.tables = dict()

    def set(self, name, data, keywords=None):
        self.tables[name] = (data, keywords)

    def table(self, name, **karg):
        return MockTable(*self.tables[name])


def test_get_info_from_ms_files():
    t = MockTables()
    t.set('a', {'TIME': np.arange(0, 16, 4), 'INTERVAL': np.ones(4) * 4}, {'SPECTRAL_WINDOW': 'SPECTRAL_WINDOW'})
    t.set('b', {'TIME': np.arange(8, 22, 4), 'INTERVAL': np.ones(4) * 4}, {'SPECTRAL_WINDOW': 'SPECTRAL_WINDOW'})
    t.set('c', {'TIME': np.arange(20, 40, 4), 'INTERVAL': np.ones(4) * 4}, {'SPECTRAL_WINDOW': 'SPECTRAL_WINDOW'})
    t.set('a/SPECTRAL_WINDOW', {'CHAN_FREQ': np.arange(100, 140, 10), 'CHAN_WIDTH': np.ones(10) * 4})
    t.set('b/SPECTRAL_WINDOW', {'CHAN_FREQ': np.arange(100, 140, 10), 'CHAN_WIDTH': np.ones(10) * 4})
    t.set('c/SPECTRAL_WINDOW', {'CHAN_FREQ': np.arange(100, 140, 10), 'CHAN_WIDTH': np.ones(10) * 4})

    with mock.patch.object(msutils, 'tables', t):
        d = msutils.get_info_from_ms_files(['a', 'b', 'c'])
        assert d['start_time'] == 0
        assert d['end_time'] == 36
        assert d['total_time'] == 12 + 12 + 16
        assert d['n_channels'] == 4
        assert d['chan_width'] == 4


def test_flag_manager(tmp_path):
    t = MockTables()
    time = np.arange(5)
    flag = np.array([0, 0, 1, 0, 1])
    freqs = np.arange(100, 140, 10)
    t.set('a', {'FLAG': flag, 'TIME': time}, {'SPECTRAL_WINDOW': 'sw'})
    t.set('b', {'FLAG': np.zeros(5), 'TIME': time}, {'SPECTRAL_WINDOW': 'sw'})
    t.set('a/SPECTRAL_WINDOW', {'CHAN_FREQ': freqs, 'CHAN_WIDTH': np.ones(10) * 4})
    t.set('b/SPECTRAL_WINDOW', {'CHAN_FREQ': freqs, 'CHAN_WIDTH': np.ones(10) * 4})

    with mock.patch.object(msutils, 'tables', t):
        fm = msutils.FlagManager.load_from_ms('a')
        assert np.allclose(fm.time, time)
        assert np.allclose(fm.flag, flag)
        assert np.allclose(fm.freqs, freqs)

        fm.save(tmp_path / 'test.h5')

        fm2 = msutils.FlagManager.load(tmp_path / 'test.h5')
        assert np.allclose(fm2.time, time)
        assert np.allclose(fm2.flag, flag)
        assert np.allclose(fm2.freqs, freqs)

        fm2.save_to_ms('b')

        fm3 = msutils.FlagManager.load_from_ms('b')
        assert np.allclose(fm3.time, time)
        assert np.allclose(fm3.flag, flag)
        assert np.allclose(fm3.freqs, freqs)


def test_slice_dim():
    l = np.random.randn(100, 20, 5)
    assert np.allclose(msutils.slice_dim(l, 1, slice(1, 3)), l[:, 1:3])
    assert np.allclose(msutils.slice_dim(l, 1, slice(1, None)), l[:, 1:])
    assert np.allclose(msutils.slice_dim(l, 0, slice(None, 20)), l[:20, :])
    assert np.allclose(msutils.slice_dim(l, 2, slice(None, -1)), l[:, :, :-1])



import os

import numpy as np

import astropy.stats as astats

from casacore import tables

from .msutils import mean_consecutive, diff_consecutive, MsDataCube, FlagManager

np.seterr(divide='ignore', invalid='ignore')


def ssins_flagger(m_vis, ntime_avg, config):
    m_vis_sum, m_vis_sum_n = mean_consecutive(diff_consecutive(m_vis, axis=1), axis=1, n=ntime_avg, return_n=True)
    ss_spc = np.mean(abs((m_vis_sum * np.sqrt(m_vis_sum_n))) ** 2, axis=(2, 3))
    for nsigma in config["nsigmas"][str(ntime_avg)]:
        ss_spc = ss_spc - np.ma.median(ss_spc, axis=1)[:, None]
        ss_spc = astats.sigma_clip(ss_spc, sigma=nsigma, stdfunc=astats.mad_std)

    mask = np.repeat(ss_spc.mask, 2 * ntime_avg, axis=1)[:, :m_vis.shape[1]]
    if mask.shape[1] < m_vis.shape[1]:
        mask = np.concatenate([mask, mask[:, -2:-1]], axis=1)

    mask[mask.sum(axis=1) > config["percentage_freq_full_flag"] * mask.shape[1]] = 1
    n_freq_full_flag = mask.shape[0] - mask.sum(axis=0).max()
    mask[:, mask.sum(axis=0) - n_freq_full_flag > config["percentage_time_full_flag"] *
         (mask.shape[0] - n_freq_full_flag)] = 1

    percentile_flag = 100. * mask.sum() / mask.size
    print(f'-> SSINS with {ntime_avg} time averaging: {percentile_flag:.2f} % time/freq flagged')

    return (np.zeros_like(m_vis.mask) + mask[:, :, None, None]).astype(bool)


def time_freq_threshold_flagger(mask, threshold):
    mask_time_freq = (1. * mask.sum(axis=(2, 3)) / (mask.shape[2] * mask.shape[3]) > threshold)
    percentile_flag = 100. * mask_time_freq.sum() / mask_time_freq.size
    print(f'-> Time/freq flagging with threshold {threshold}: {percentile_flag:.2f} % time/freq flagged')

    return (np.zeros_like(mask) + mask_time_freq[:, :, None, None]).astype(bool)


def baseline_threshold_flagger(mask, threshold):
    idx_freq_full_flag = (mask[:, :, :, 0].sum(axis=(1, 2)) == mask.shape[1] * mask.shape[2])

    mask_baseline = (1. * mask[~idx_freq_full_flag].sum(axis=(0, 1, 3)) /
                     (mask[~idx_freq_full_flag].shape[0] * mask.shape[1] * mask.shape[3]) > threshold)
    percentile_flag = 100. * mask_baseline.sum() / mask_baseline.size
    print(f'-> Baseline flagging with threshold {threshold}: {percentile_flag:.2f} % time/freq flagged')

    return (np.zeros_like(mask) + mask_baseline[None, None, :, None]).astype(bool)


def snapshot_threshold_flagger(mask, threshold):
    idx_freq_full_flag = (mask[:, :, :, 0].sum(axis=(1, 2)) == mask.shape[1] * mask.shape[2])
    if mask[~idx_freq_full_flag].sum() / float(mask[~idx_freq_full_flag].size) > threshold:
        print(f'-> Fully flag the MS as the flag ratio is above threshold of {threshold}')
        return np.ones_like(mask)
    return mask


def get_ss_spc(m_vis, zero_mean=False):
    ss_spc = np.mean(abs(diff_consecutive(m_vis, axis=1)) ** 2, axis=(2, 3))
    if zero_mean:
        ss_spc = ss_spc - np.ma.median(ss_spc, axis=1)[:, None]

    return ss_spc


def get_badstatsions(ms, nsigma):
    file = ms + '/QUALITY_BASELINE_STATISTIC'
    if not os.path.exists(file):
        print('Error: QUALITY_BASELINE_STATISTIC does not exits. Did you run aoquality ?')
        return []

    with tables.table(file, ack=False) as t:
        kind = t.getcol('KIND')
        a1 = t.getcol('ANTENNA1')
        a2 = t.getcol('ANTENNA2')
        v = t.getcol('VALUE')

    a_max = max(max(a1), max(a2))
    idx = kind == 7
    i = abs(v.mean(axis=1)[idx])
    aa1 = a1[idx]
    aa2 = a2[idx]
    i[aa1 == aa2] = np.nan
    stat = np.nanmedian(np.array([i[(aa1 == k) | (aa2 == k)] for k in range(a_max)]), axis=0)
    bad_statsions = np.where(astats.sigma_clip(stat, sigma=nsigma).mask > 0)[0]

    return bad_statsions


def get_badbaselines(ms, nsigma_station, nsigma_baseline):
    file = ms + '/QUALITY_BASELINE_STATISTIC'
    if not os.path.exists(file):
        print('Error: QUALITY_BASELINE_STATISTIC does not exits. Did you run aoquality ?')
        return []

    with tables.table(file, ack=False) as t:
        kind = t.getcol('KIND')
        a1 = t.getcol('ANTENNA1')
        a2 = t.getcol('ANTENNA2')
        v = t.getcol('VALUE')

    a_max = max(max(a1), max(a2))
    idx = (kind == 7)
    i = np.ma.array(abs(v[idx].mean(axis=1)))
    a1 = a1[idx]
    a2 = a2[idx]
    i[a1 == a2] = np.ma.masked
    i[i == 0] = np.ma.masked
    stat = np.median(np.array([i[(a1 == k) | (a2 == k)] for k in range(a_max)]), axis=0)
    badstations = np.where(astats.sigma_clip(stat, sigma=nsigma_station).mask > 0)[0].tolist()
    i[np.sum([(a1 == k) | (a2 == k) for k in badstations], axis=0).astype(bool) > 0] = np.ma.masked
    mask_before = i.mask
    i = astats.sigma_clip(i, sigma=nsigma_baseline)
    badbaselines = ['%s&%s' % (k, v) for k, v in zip(a1[(i.mask ^ mask_before)], a2[(i.mask ^ mask_before)])]

    return ';'.join([str(k) for k in badstations] + badbaselines)


def flag_badscans(mss, umin=20, umax=400, data_col='CORRECTED_DATA', nsigma=5):
    all_var = np.array([np.var(MsDataCube.load(ms, umin, umax, data_col, verbose=False).data_dt,
                               axis=(0, 1, 2)) for ms in mss])
    bad_scans = astats.sigma_clip(all_var, sigma=nsigma, axis=0, stdfunc=astats.mad_std).mask.sum(axis=1) > 0
    print('Flagging bad scans ...')
    for ms in np.array(mss)[bad_scans]:
        FlagManager.fully_flag(ms)
